import torch
import math

def cosine_schedule(base: float, end: float, T: int, warmup: int = 0, flat_end: int = 0):
    method_T = T - warmup - flat_end
    method_vals = end - (end - base) * ((torch.arange(0, method_T, 1) * math.pi / method_T).cos() + 1) / 2
    return build_schedule(method_vals, T, base, end, warmup, flat_end)

def step_schedule(base: float, T: int, milestones: list[int], gamma: float, warmup: int = 0, flat_end: int = 0):
    method_T = T - warmup - flat_end
    method_vals, val, t = torch.tensor([]), base, 0

    for milestone in milestones:
        method_vals = torch.cat([method_vals, torch.ones(milestone - t) * val])
        t = milestone
        val = val * gamma

    if t < method_T:
        method_vals = torch.cat([method_vals, torch.ones(method_T - t) * val])

    return build_schedule(method_vals, T, base, None, warmup, flat_end, milestones, gamma)

def flat_schedule(base: float, T: int, warmup: int = 0, flat_end: int = 0):
    method_T = T - warmup - flat_end
    method_vals = torch.ones(method_T) * base
    return build_schedule(method_vals, T, base, None, warmup, flat_end)

def build_schedule(method_vals: torch.Tensor, T: int, base: float, end: float = None, warmup: int = 0, flat_end: float = 0, milestones: list[int] = None, gamma: float = None):

    # Linear warmup phase, from 0 to base at start of optimisation
    warmup_vals = torch.linspace(0.0, base, warmup+1)[1:] if warmup > 0 else torch.tensor([])

    # Flat end phase, keep the hyperparameter constant at the end
    flat_end_vals = torch.ones(flat_end) * method_vals[-1] if flat_end > 0 else torch.tensor([])

    return torch.cat([warmup_vals, method_vals, flat_end_vals])


import torch
from joetorch.nn.modules import MLP, EncBlock, DecBlock
import torch.nn.functional as F

def MNIST_MLP_Encoder(out_dim: int, layer_norm: bool=True):
    out_actv_fn = torch.nn.LayerNorm(out_dim) if layer_norm else None
    modules =[
        torch.nn.Flatten(),
        MLP(in_dim=784, out_dim=out_dim, num_hidden=3, hidden_dim=512, hidden_actv_module=torch.nn.SiLU(), out_actv_module=out_actv_fn)
    ]
    if layer_norm:
        modules.append(torch.nn.LayerNorm(out_dim))
    return torch.nn.Sequential(*modules)

def MNIST_MLP_Decoder(in_dim: int):
    modules = [
        MLP(in_dim=in_dim, out_dim=784, num_hidden=3, hidden_dim=512, hidden_actv_module=torch.nn.SiLU(), out_actv_module=torch.nn.Sigmoid()),
        torch.nn.Unflatten(1, (1, 28, 28))
    ]
    return torch.nn.Sequential(*modules)

class MNIST_CNN_Encoder(torch.nn.Module):
    def __init__(self, out_dim: int, mlp: bool=True, layer_norm: bool=True):
        super().__init__()
        modules = [torch.nn.Sequential(
            EncBlock(1, 64, 3, 1, 1, pool=True),
            EncBlock(64, 128, 3, 1, 1, pool=True),
            EncBlock(128, 256, 3, 1, 0),
            EncBlock(256, 512, 3, 1, 0),
            EncBlock(512, out_dim, 3, 1, 0),
            torch.nn.Flatten(),
        )]
        if mlp:
            modules.append(MLP(in_dim=out_dim, out_dim=out_dim, num_hidden=2, hidden_dim=512, hidden_actv_module=torch.nn.SiLU(), out_actv_module=None))
        if layer_norm:
            modules.append(torch.nn.LayerNorm(out_dim))
        self.net = torch.nn.Sequential(*modules)

    def forward(self, x):
        return self.net(x)


class MNIST_CNN_Decoder(torch.nn.Module):
    def __init__(self, in_dim: int):
        super().__init__()
        self.net = torch.nn.Sequential(
            torch.nn.Unflatten(1, (in_dim, 1, 1)),
            DecBlock(in_dim, 512, 3, 1),
            DecBlock(512, 256, 3, 3),
            DecBlock(256, 128, 3, 3),
            DecBlock(128, 64, 2, 1),
            torch.nn.Conv2d(64, 1, 3, 1, 1),
        )

    def forward(self, x):
        return self.net(x)

class MNIST_AE(torch.nn.Module):
    def __init__(self, out_dim: int, mode: str, layer_norm: bool=True):
        super().__init__()
        if mode == 'mlp':
            self.encoder = MNIST_MLP_Encoder(out_dim, layer_norm)
            self.decoder = MNIST_MLP_Decoder(out_dim)
        elif mode == 'cnn':
            self.encoder = MNIST_CNN_Encoder(out_dim, layer_norm)
            self.decoder = MNIST_CNN_Decoder(out_dim)
        else:
            raise ValueError(f'Invalid mode: {mode}')
    
    def infer(self, x):
        return self.encoder(x)

    def forward(self, x):
        return self.decoder(self.encoder(x))
    
    def loss(self, batch, **kwargs):
        x, _ = batch
        x_hat = self.forward(x)
        mse = F.mse_loss(x_hat, x)
        return {'loss': mse, 'mse': mse}


from joetorch.datasets.mnist import MNIST, FashionMNIST
from joetorch.datasets.dataset import PreloadedDataset
from joetorch.datasets.imagenet import ImageNet

__all__ = ['MNIST', 'FashionMNIST', 'PreloadedDataset', 'ImageNet']


from joetorch.nn.modules.fully_connected import MLP
from joetorch.nn.modules.convolutional import EncBlock, DecBlock, ConvResBlock, ConvMixer

__all__ = ['MLP', 'EncBlock', 'DecBlock', 'ConvResBlock', 'ConvMixer']



import torch
import torch.nn as nn
import torchvision.transforms as transforms
import torchvision.datasets as datasets
from torch.utils.data import Dataset
import os
import numpy as np
import random
from PIL import Image
import xml.etree.ElementTree as ET

from joetorch.datasets.dataset import PreloadedDataset

class ImageNet(Dataset):
    def __init__(self, root, split, subset_size=None, size=(224, 224)):
        self.root = root
        self.split = split
        assert split in ['train', 'val', 'test']

        self.data_dir = os.path.join(root, f'ILSVRC/Data/CLS-LOC/{split}')
        self.to_tensor = transforms.ToTensor()
        self.size = size

        self.folder_to_label = {folder: i for i, folder in enumerate(os.listdir(os.path.join(root, 'ILSVRC/Data/CLS-LOC/train')))}

        if split == 'train':
            self.transform = transforms.Compose([
                transforms.RandomResizedCrop(224),  # Random crop & resize
                transforms.RandomHorizontalFlip(),  # Augmentation
                transforms.ToTensor(),              # Convert to tensor
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),  # Standard ImageNet normalization
            ])
        elif split == 'val' or split == 'test':
            self.transform = transforms.Compose([
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            ])

        self.image_paths = []
        self.labels = []
        if split == 'train':
            subdirs = [os.path.join(self.data_dir, d) for d in os.listdir(self.data_dir) 
                       if os.path.isdir(os.path.join(self.data_dir, d))]

            for d in subdirs:
                class_name = d.split('/')[-1]
                for fname in os.listdir(d):
                    if fname.lower().endswith(('.png', '.jpg', '.jpeg')):
                        self.image_paths.append(os.path.join(d, fname))
                        self.labels.append(self.folder_to_label[class_name])
        else:
            label_dir = os.path.join(root, f'ILSVRC/Annotations/CLS-LOC/{split}')
            for fname in os.listdir(self.data_dir):
                if fname.lower().endswith(('.png', '.jpg', '.jpeg')):
                    self.image_paths.append(os.path.join(self.data_dir, fname))
                    annotation = os.path.join(label_dir, fname.replace('.JPEG', '.xml'))
                    tree = ET.parse(annotation)
                    root = tree.getroot()
                    for obj in root.findall('object'):
                        label = obj.find('name').text
                        self.labels.append(self.folder_to_label[label])

        if subset_size is not None:
            indices = np.random.choice(len(self.image_paths), subset_size, replace=False)
            self.image_paths = [self.image_paths[i] for i in indices]

        self.image_paths.sort()

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        img_path = self.image_paths[idx]
        image = Image.open(img_path).convert('RGB')
        image = self.transform(image)

        label = self.labels[idx]

        return image, label


import torch
import torch.nn.functional as F

def feature_std(x):
    """
    Compute the standard deviation of the features.
    Args:
        x: torch.Tensor, the input tensor.
    Returns:
        torch.Tensor, the computed standard deviation.
    """
    # x: (N, C)
    return x.std(dim=0).mean()

def feature_std_entropy(x):
    """
    Compute the entropy of the standard deviations of the features.
    Args:
        x: torch.Tensor, the input tensor.
    Returns:
        torch.Tensor, the computed entropy.
    """
    # x: (N, C)
    stds = x.std(dim=0) # stds: (C,)
    normalised = torch.softmax(stds, dim=0) # normalised: (C,)
    entropy = -torch.sum(normalised * torch.log(normalised + 1e-8)) # entropy: ([])
    return entropy

def feature_correlation_matrix(x, standardise:bool=True, unbiased:bool=True):
    """
    Compute the correlation matrix of the features.
    Args:
        x: torch.Tensor, the input tensor.
        standardise: bool, whether to standardise the features.
        unbiased: bool, whether to use the unbiased estimation of the correlation matrix. Use unbiased=True when evaluating encoder or population, and unbiased=False when evaluating the batch.
    Returns:
        torch.Tensor, the computed correlation matrix.
    """
    # x: (N, C)
    if x.size(0) <= 1:
        return torch.tensor(0.0, device=x.device)
    
    # Standardise features, so each feature has mean 0 and std 1
    if standardise:
        mean = x.mean(dim=0, keepdim=True)
        std = x.std(dim=0, keepdim=True, unbiased=True)
        x = (x - mean) / (std + 1e-8)
    
    # Compute normalised auto-correlation
    # Using unbiased estimation (N-1)
    return torch.matmul(x.T, x) / (x.size(0) - 1)

def interfeature_correlation(x, standardise:bool=True, unbiased:bool=True, return_matrix:bool=False):
    """
    Compute the average absolute interfeature correlation of the given input tensor, ignoring self-correlation (where i == j).
    Args:
        x: torch.Tensor, the input tensor.
        standardise: bool, whether to standardise the features.
        unbiased: bool, whether to use the unbiased estimation of the correlation matrix. Use unbiased=True when evaluating encoder or population, and unbiased=False when evaluating the batch.
        return_matrix: bool, whether to return the correlation matrix.
    Returns:
        torch.Tensor, the computed interfeature correlation.
    """

    corr_matrix = feature_correlation_matrix(x, standardise, unbiased)
    
    # Take absolute values since we care about correlation strength
    # regardless of direction
    corrs = corr_matrix.abs()
    
    # select above diagonal to avoid double counting and self-correlation
    corrs = torch.triu(corrs, diagonal=1)
    
    # If there's only one feature, return 0
    if x.size(1) <= 1:
        return torch.tensor(0.0, device=x.device)
    
    # average, accounting for the number of elements above diagonal
    n_elements = (x.size(1) * (x.size(1) - 1)) // 2
    if return_matrix:
        return corrs, corr_matrix
    else:
        return corrs.sum() / n_elements

def get_representations(model, dataset, amp:bool=True):
    device = next(model.parameters()).device
    batch_size = 256
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=False)

    embeddings = []
    labels = []

    for x, y in dataloader:
        x = x.to(device)
        with torch.amp.autocast(device.type, enabled=amp, dtype=torch.bfloat16):
            emb = model.embed(x)
        embeddings.append(emb.detach())
        labels.append(y)
    
    embeddings = torch.cat(embeddings)
    labels = torch.cat(labels)
    
    return embeddings, labels